@@ -171,12 +171,30 @@ def create_model(use_my_model, inputs, meta_outputs, keras_options, var_df,
     val_monitor = keras_options["monitor"]
     val_loss = keras_options["loss"]
     val_metrics = keras_options["metrics"]
-    if not keras_options['optimizer']:
-        ### if it is an empty string, use the default Adam optimizer
-        optimizer = Adam(lr=0.01, beta_1=0.9, beta_2=0.999)
-    else:
+    learning_rate = 5e-2
+    ############################################################################
+    try:
+        print('    number of outputs = %s, output_activation = %s' %(
+                            num_labels, output_activation))
+        print('    loss function: %s' %str(val_loss).split(".")[-1].split(" ")[0])
+    except:
+        print('    loss fn = %s    number of outputs = %s, output_activation = %s' %(
+                            val_loss, num_labels, output_activation))
+    try:
         optimizer = return_optimizer(keras_options['optimizer'])
-
+    except:
+        #####   set some default optimizers here for model parameters here ##
+        if not keras_options['optimizer']:
+            optimizer = keras.optimizers.SGD(learning_rate)
+        elif keras_options["optimizer"] in ['RMS', 'RMSprop']:
+            optimizer = keras.optimizers.RMSprop(learning_rate)
+        elif keras_options['optimizer'] in ['Adam', 'adam', 'ADAM', 'NADAM', 'Nadam']:
+            optimizer = keras.optimizers.Adam(learning_rate)
+        else:
+            optimizer = keras.optimizers.Adagrad(learning_rate)
+    keras_options['optimizer'] = optimizer
+    print('    initial learning rate = %s' %learning_rate)
+    print('    initial optimizer = %s' %str(optimizer).split(".")[-1].split(" ")[0])
     ###################################################################################
     if data_dim <= 1e6:
         dense_layer1 = max(96,int(data_dim/30000))
@@ -197,17 +215,17 @@ def create_model(use_my_model, inputs, meta_outputs, keras_options, var_df,
                                 dense_layer1,dense_layer2,dense_layer3))
     fast_models = ['deep_and_wide','deep_wide','wide_deep', 
                                 'wide_and_deep','deep wide', 'wide deep', 'fast', 'fast1']
-    
+    fast_models2 = ['deep_and_cross', 'deep_cross', 'deep cross', 'fast2']
     #### The Deep and Wide Model is a bit more complicated. So it needs some changes in inputs! ######
     prebuilt_models = ['basic', 'simple', 'default','simple_dnn','sample model',
                         'deep', 'big_deep', 'big deep', 'giant_deep', 'giant deep',
                         'cnn1', 'cnn','cnn2'] 
     ######   Just do a simple check for auto models here ####################
-    if keras_model_type.lower() in fast_models+prebuilt_models:
-            all_inputs = copy.deepcopy(inputs)
+    if keras_model_type.lower() in fast_models+prebuilt_models+fast_models2:
+            all_inputs = inputs
     else:
         ### this means it's an auto model and you create one here 
-        print('    building a %s model...' %keras_model_type)
+        print('    creating %s model body...' %keras_model_type)
         num_layers = check_keras_options(keras_options, 'num_layers', 1)
         model_body = tf.keras.Sequential([])
         for l_ in range(num_layers):
@@ -271,7 +289,7 @@ def create_model(use_my_model, inputs, meta_outputs, keras_options, var_df,
                 #final_outputs = layers.Dense(units=num_predicts, activation=output_activation)(merged)
                 #model_body = keras.Model(inputs=all_inputs, outputs=final_outputs)
                 print('    Created deep and wide %s model, ...' %keras_model_type)
-            elif keras_model_type.lower() in ['deep_and_cross', 'deep_cross', 'deep cross', 'fast2']:
+            elif keras_model_type.lower() in fast_models2:
                 dropout_rate = 0.1
                 #hidden_units = [32, 32]
                 hidden_units = [dense_layer1, dense_layer2]
