@@ -360,10 +360,9 @@ class MyTuner(Tuner):
             #scores.append(score)
         ##### This is where we capture the best learning rate from the optimizer chosen ######
         model_lr = comp_model.optimizer.learning_rate.numpy()
-        
-        self.user_var = model_lr
-        #print('    model lr = %s' %model_lr)
-        trial.metrics['final_lr'] = history.history['lr'][-1]
+        #self.user_var = model_lr
+        print('    found best learning rate = %s' %model_lr)
+        trial.metrics['final_lr'] = model_lr
         #print('    trial final_lr = %s' %trial.metrics['final_lr'])
         self.score_trial(trial, score) 
         #self.score_trial(trial, min(scores)) 
@@ -383,25 +382,27 @@ def return_optimizer_trials(hp, hpq_optimizer):
     nadam = keras.optimizers.Nadam(lr=0.001, beta_1=0.9, beta_2=0.999)
     best_optimizer = ''
     #############################################################################
-    if hpq_optimizer == 'Adam':
+    if hpq_optimizer.lower() in ['adam']:
         best_optimizer = tf.keras.optimizers.Adam(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]),
             epsilon=hp.Param('epsilon', [1e-6, 1e-8, 1e-10, 1e-12, 1e-14], ordered=True))
-    elif hpq_optimizer == 'SGD':
+    elif hpq_optimizer.lower() in ['sgd']:
         best_optimizer = keras.optimizers.SGD(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]),
                              momentum=0.9)
-    elif hpq_optimizer == 'Nadam':
+    elif hpq_optimizer.lower() in ['nadam']:
         best_optimizer = keras.optimizers.Nadam(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]),
                          beta_1=0.9, beta_2=0.999)
-    elif hpq_optimizer == 'AdaMax':
+    elif hpq_optimizer.lower() in ['adamax']:
         best_optimizer = keras.optimizers.Adamax(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]),
                          beta_1=0.9, beta_2=0.999)
-    elif hpq_optimizer == 'Adagrad':
+    elif hpq_optimizer.lower() in 'adagrad':
         best_optimizer = keras.optimizers.Adagrad(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]))
-    elif hpq_optimizer == 'RMSprop':
+    elif hpq_optimizer.lower() in ['rmsprop']:
         best_optimizer = keras.optimizers.RMSprop(lr=hp.Param('init_lr', [1e-2, 1e-3, 1e-4]),
                          rho=0.9)
-    else:
+    elif hpq_optimizer.lower() in ['nesterov']:
         best_optimizer = keras.optimizers.SGD(lr=0.001, momentum=0.9, nesterov=True)
+    else:
+        best_optimizer = keras.optimizers.SGD(lr=0.001, momentum=0.9)
     return best_optimizer
 #####################################################################################
 def return_optimizer(hpq_optimizer):
@@ -409,14 +410,15 @@ def return_optimizer(hpq_optimizer):
     This returns the keras optimizer with proper inputs if you send the string.
     hpq_optimizer: input string that stands for an optimizer such as "Adam", etc.
     """
+    learning_rate_set = 5e-2
     ##### These are the various optimizers we use ################################
-    momentum = keras.optimizers.SGD(lr=0.001, momentum=0.9)
-    nesterov = keras.optimizers.SGD(lr=0.001, momentum=0.9, nesterov=True)
-    adagrad = keras.optimizers.Adagrad(lr=0.001)
-    rmsprop = keras.optimizers.RMSprop(lr=0.001, rho=0.9)
-    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999)
-    adamax = keras.optimizers.Adamax(lr=0.001, beta_1=0.9, beta_2=0.999)
-    nadam = keras.optimizers.Nadam(lr=0.001, beta_1=0.9, beta_2=0.999)
+    momentum = keras.optimizers.SGD(lr=learning_rate_set, momentum=0.9)
+    nesterov = keras.optimizers.SGD(lr=learning_rate_set, momentum=0.9, nesterov=True)
+    adagrad = keras.optimizers.Adagrad(lr=learning_rate_set)
+    rmsprop = keras.optimizers.RMSprop(lr=learning_rate_set, rho=0.9)
+    adam = keras.optimizers.Adam(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
+    adamax = keras.optimizers.Adamax(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
+    nadam = keras.optimizers.Nadam(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
     best_optimizer = ''
     #############################################################################
     #### This could be turned into a dictionary but for now leave is as is for readability ##
@@ -469,29 +471,32 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     patience = keras_options["patience"]
     cols_len = len([item for sublist in list(var_df.values()) for item in sublist])
     data_dim = int(data_size*meta_outputs.shape[1])
+    optimizer = keras_options['optimizer']
     early_stopping = check_keras_options(keras_options, "early_stopping", False)
     print('After preprocessing using keras layers, features dimensions is now %s' %meta_outputs.shape[1])
     print('    original datasize = %s, initial batchsize = %s' %(data_size, batch_size))
     print("    Early stopping : %s" %early_stopping)
     NUMBER_OF_EPOCHS = check_keras_options(keras_options, "epochs", 100)
-    learning_rate = 5e-1
+    if keras_options['lr_scheduler'] in ['expo', 'ExponentialDecay', 'exponentialdecay']:
+        print('    chosen ExponentialDecay learning rate scheduler')
+        expo_steps = (NUMBER_OF_EPOCHS*data_size)//batch_size
+        learning_rate = keras.optimizers.schedules.ExponentialDecay(0.01, expo_steps, 0.1)
+    else:
+        learning_rate = check_keras_options(keras_options, "learning_rate", 5e-2)
+    #### The steps are actually not needed but remove them later.###
     if len(var_df['nlp_vars']) > 0:
         steps = 10
     else:
-        steps = max(50, data_size//(batch_size*25))
+        steps = max(10, (data_size//(batch_size*2)))
         steps = min(300, steps)
     print('    recommended steps per epoch = %d' %steps)
     STEPS_PER_EPOCH = check_keras_options(keras_options, "steps_per_epoch", 
                         steps)
-    #keras.optimizers.schedules.ExponentialDecay(0.01,STEPS_PER_EPOCH, 0.95)
     #### These can be standard for every keras option that you use layers ######
     kernel_initializer = check_keras_options(keras_options, 'kernel_initializer', 'lecun_normal')
     activation='selu'
     print('    default initializer = %s, default activation = %s' %(kernel_initializer, activation))
-    #####   set some defaults for model parameters here ##
-    optimizer = check_keras_options(keras_options,'optimizer', Adam(lr=learning_rate, beta_1=0.9, beta_2=0.999))
-    #optimizer = SGD(lr=learning_rate, momentum = 0.9)
-    print('    Using optimizer = %s' %str(optimizer).split(".")[-1][:8])
+    ############################################################################
     use_bias = check_keras_options(keras_options, 'use_bias', True)
     lr_scheduler = check_keras_options(keras_options, 'lr_scheduler', "")
     onecycle_steps = max(10, np.ceil(data_size/(2*batch_size))*NUMBER_OF_EPOCHS)
@@ -505,8 +510,13 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     val_loss = keras_options["loss"]
     val_metrics = keras_options["metrics"]
     ########################################################################
-    print('    loss fn = %s\n    num predicts = %s, output_activation = %s' %(
-                        val_loss, num_predicts, output_activation))
+    try:
+        print('    number of classes = %s, output_activation = %s' %(
+                            num_predicts, output_activation))
+        print('    loss function: %s' %str(val_loss).split(".")[-1].split(" ")[0])
+    except:
+        print('    loss fn = %s    number of classes = %s, output_activation = %s' %(
+                            val_loss, num_predicts, output_activation))
     ####  just use modeltype for printing that's all ###
     modeltype = cat_vocab_dict['modeltype']
 
@@ -517,11 +527,10 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     print('    steps_per_epoch = %s, number epochs = %s' %(STEPS_PER_EPOCH, NUMBER_OF_EPOCHS))
     print('    val mode = %s, val monitor = %s, patience = %s' %(val_mode, val_monitor, patience))
 
-    onecycle_steps = math.ceil(STEPS_PER_EPOCH * NUMBER_OF_EPOCHS)
-    print('    recommended OneCycle steps = %d' %onecycle_steps)
-
     callbacks_dict, tb_logpath = get_callbacks(val_mode, val_monitor, patience,
                                     learning_rate, save_weights_only, onecycle_steps)
+    chosen_callback = get_chosen_callback(callbacks_dict, keras_options)
+    print('    chosen keras LR scheduler = %s' %chosen_callback)
 
     ###### You can use Storm Tuner to set the batch size ############
     ############## Split train into train and validation datasets here ###############
@@ -569,7 +578,7 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
         tuner = "None"
     else:
         tuner = model_options["tuner"]
-    print('Training %s model using %s. This will take time...' %(keras_model_type, tuner))
+    print('    Training %s model using %s. This will take time...' %(keras_model_type, tuner))
 
     from secrets import randbelow
     rand_num = randbelow(10000)
@@ -582,12 +591,20 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     #######################################################################################
     ###    E A R L Y    S T O P P I N G    T O    P R E V E N T   O V E R F I T T I N G  ##
     #######################################################################################
-    chosen_callback = get_chosen_callback(callbacks_dict, keras_options)
-    callbacks_list_tuner = [chosen_callback, callbacks_dict['es']]
+    if keras_options['lr_scheduler'] in ['expo', 'ExponentialDecay', 'exponentialdecay']:
+        callbacks_list_tuner = callbacks_dict['es']
+    else:
+        callbacks_list_tuner = [chosen_callback, callbacks_dict['es']]
+
     targets = cat_vocab_dict["target_variables"]
     ############################################################################
     ########     P E R FO R M     T U N I N G    H E R E  ######################
     ############################################################################
+    tune_mode = 'min'
+    if num_labels > 1 and modeltype != 'Regression':
+        tune_mode = 'max'
+    else:
+        tune_mode = val_mode
     if tuner.lower() == "storm":
         trials_saved_path = os.path.join(project_name,keras_model_type)
         if not os.path.exists(trials_saved_path):
@@ -596,7 +613,7 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
         randomization_factor = 0.50
         tuner = MyTuner(project_dir=trials_saved_path,
                     build_fn=build_model_storm,
-                    objective_direction=val_mode,
+                    objective_direction=tune_mode,
                     init_random=5,
                     max_iters=max_trials,
                     randomize_axis_factor=randomization_factor,
@@ -632,8 +649,11 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
             best_model, best_optimizer = build_model_storm(hpq, batch_size)
             best_batch = hpq.values['batch_size']
             hpq_optimizer = hpq.values['optimizer']
-            optimizer_lr = best_trial.metrics['final_lr']
-            print('\nBest optimizer = %s and best learning_rate = %s' %(hpq_optimizer, optimizer_lr))
+            if best_trial.metrics['final_lr'] < 0:
+                print('    best learning rate less than zero. Resetting it....')
+                optimizer_lr = 0.01
+            else:
+                optimizer_lr = best_trial.metrics['final_lr']
             print('Best hyperparameters: %s' %hpq.values)
         except:
             ### Sometimes the tuner cannot find a config that works!
@@ -647,9 +667,7 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
 
         ### Sometimes the learning rate is below zero - so reset it here!
         ### Set the learning rate for the best optimizer here ######
-        if optimizer_lr < 0:
-            print('    best learning rate less than zero. Resetting it....')
-            optimizer_lr = 0.01
+        print('\nSetting best optimizer %s its best learning_rate = %s' %(hpq_optimizer, optimizer_lr))
         K.set_value(best_optimizer.learning_rate, optimizer_lr)
 
         ##### This is the simplest way to convert a sequential model to functional model!
@@ -694,7 +712,7 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
             return score
         ##### This where you run optuna ###################
         study_name = project_name+'_'+keras_model_type+'_study_'+str(rand_num)
-        if val_mode == 'max':
+        if tune_mode == 'max':
             study = optuna.create_study(study_name=study_name, direction="maximize", load_if_exists=False)
         else:
             study = optuna.create_study(study_name=study_name, direction="minimize", load_if_exists=False)
@@ -730,15 +748,10 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
                             num_labels, best_optimizer, val_loss, val_metrics, cols_len, targets)
         deep_model = get_compiled_model(inputs, deep_outputs, output_activation, num_predicts, 
                             num_labels, best_optimizer, val_loss, val_metrics, cols_len, targets)
-    #######################################################################################
-    #### here we can define the custom logic to assign a score to the model to monitor
-    if num_labels > 1:
-        ### You must choose one of the label outputs to monitor - we will choose the last one
-        val_monitor = val_monitor.split("_")[-1]
-        val_monitor = 'val_output_'+str(num_labels-1)+'_' + val_monitor
+        #######################################################################################
 
     ####################################################################################
-    #####   T E N S O R  B O A R D    C  A  N     B E   F O U N D    H E R E ######
+    #####   T R A IN  A N D   V A L I D A T I O N   F O U N D    H E R E          ######
     ####################################################################################
     
     #train_ds = train_ds.unbatch().batch(best_batch)
@@ -752,24 +765,31 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     ############### F I R S T  T R A I N   F O R  1 0 0   E P O C H S ##################
     ### You have to set both callbacks in order to learn what the best learning rate is 
     ####################################################################################
-    np.random.seed(42)
-    tf.random.set_seed(42)
-
-    if early_stopping:
-        callbacks_list = [chosen_callback, callbacks_dict['es'], callbacks_dict['tb']]
+    if keras_options['lr_scheduler'] in ['expo', 'ExponentialDecay', 'exponentialdecay']:
+        if early_stopping:
+            callbacks_list = [callbacks_dict['es'], callbacks_dict['tb']]
+        else:
+            callbacks_list = [callbacks_dict['tb']]
     else:
-        callbacks_list = [chosen_callback, callbacks_dict['tb']]
+        if early_stopping:
+            callbacks_list = [callbacks_dict['es'], callbacks_dict['tb'], callbacks_dict['rlr']]
+        else:
+            callbacks_list = [callbacks_dict['tb'], callbacks_dict['rlr']]
 
     print('Model training with best hyperparameters for %d epochs' %NUMBER_OF_EPOCHS)
     for each_callback in callbacks_list:
         print('    Callback added: %s' %str(each_callback).split(".")[-1])
     
-    ####  Do this with LR scheduling but NO early stopping since we want the model fully trained #####
+    ############################    M O D E L     T R A I N I N G   ##################
+    np.random.seed(42)
+    tf.random.set_seed(42)
     history = best_model.fit(train_ds, validation_data=valid_ds, batch_size=best_batch,
             epochs=NUMBER_OF_EPOCHS, #steps_per_epoch=STEPS_PER_EPOCH, 
             callbacks=callbacks_list, class_weight=class_weights,
             #validation_steps=STEPS_PER_EPOCH, 
            shuffle=True)
+    print('    Model training completed. Following metrics available: %s' %history.history.keys())
+    print('Time taken to train model (in mins) = %0.0f' %((time.time()-start_time)/60))
 
     #################################################################################
     #######          R E S E T    K E R A S      S E S S I O N 
@@ -782,11 +802,12 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
     tf.keras.backend.reset_uids()
     
     ###   Once the best learning rate is chosen the model is ready to be trained on full data
-    print('    Model training metrics available: %s' %history.history.keys())
     try:
         stopped_epoch = int(pd.DataFrame(history.history).shape[0] - patience) ## this is where it stopped
     except:
         stopped_epoch = 100
+    print('    Stopped epoch = %s' %stopped_epoch)
+
     ###  Plot the epochs and loss metrics here #####################
     try:
         if modeltype == 'Regression':
@@ -922,8 +943,13 @@ def train_custom_model(inputs, meta_outputs, full_ds, target, keras_model_type,
         if isinstance(target, str):
             plt.figure(figsize=(15,6))
             ax1 = plt.subplot(1, 2, 1)
-            ax1.scatter(x=y_test, y=y_test_preds,)
-            ax1.set_title('Actuals (x-axis) vs. Predictions (y-axis)')
+            residual = np.sqrt((y_test - y_test_preds)**2)
+            lowermin = min(y_test.min(), residual.min())
+            uppermax = max(y_test.max(), residual.max())
+            ax1.scatter(x=y_test, y=residual)
+            ax1.set_xlim(lowermin, uppermax)
+            ax1.set_ylim(lowermin, uppermax)
+            ax1.set_title('Actuals (x-axis) vs. RMSE of predictions (y-axis)')
             pdf = save_valid_predictions(y_test, y_test_preds.ravel(), project_name, num_labels)
             ax2 = plt.subplot(1, 2, 2)
             pdf.plot(ax=ax2)
